import numpy as np
import pandas as pd
import pandas_ta as pta
import freqtrade.vendor.qtpylib.indicators as qtpylib
from freqtrade.strategy import IStrategy, stoploss_from_absolute
from freqtrade.persistence import Trade
from datetime import datetime


class newIchimokuStrategy(IStrategy):

    # Timeframe
    timeframe = '15m'

    # ROI table:
    minimal_roi = {
        "0": 1.0,
        "1000": 0
    }

    # Stoploss:
    stoploss = -0.2

    # Other strategy flags
    process_only_new_candles = True
    use_exit_signal = True
    can_short = False
    use_custom_stoploss = True
    exit_profit_only = True


    def custom_stoploss(self, pair: str, Trade: 'Trade', current_profit: float, current_rate: float, **kwargs) -> float:
        
        # Retrieves the analyzed dataframe for the specified currency pair and timeframe
        dataframe, _ = self.dp.get_analyzed_dataframe(pair, self.timeframe)
        # Extracts the last candle (most recent data point) from the dataframe
        last_candle = dataframe.iloc[-1].squeeze()

        if Trade.stop_loss == Trade.initial_stop_loss:
            # If current stoploss is same as the initial stoploss value,
            # update it based on the lowest span line for the last candle
            spanline_stoploss = stoploss_from_absolute(last_candle['stoploss_line'], current_rate)
            if np.isnan(spanline_stoploss):
                return None
            else:
                return spanline_stoploss

        # Use the initial stoploss until the profit is above 4%,
        # then use a trailing stoploss of 50% of the current profit with a minimum of 2.5% and a maximum of 5%.
        if current_profit < 0.04:
            return -1 # return a value bigger than the initial stoploss to keep using the initial stoploss

        # After reaching the desired offset, allow the stoploss to trail by half the profit
        desired_stoploss = current_profit / 2

        # Use a minimum of 2.5% and a maximum of 5%
        return max(min(desired_stoploss, 0.05), 0.025)
    
    def custom_exit(self, pair: str, trade: 'Trade', current_time: 'datetime', current_rate: float, current_profit: float, **kwargs):
        
        # Retrieves the analyzed dataframe for the specified currency pair and timeframe
        dataframe, _ = self.dp.get_analyzed_dataframe(pair, self.timeframe)
        # Extracts the last candle (most recent data point) from the dataframe
        last_candle = dataframe.iloc[-1].squeeze()
        
        # Calculates risk, target, risk reward ratio
        risk = trade.open_rate - last_candle['stoploss_line']
        target = risk * 2
        # risk_reward_ratio = target / risk
        
        dataframe.loc[
            (
                # Exit trade when price hits target
                (current_profit >= target)
            ),
            'exit_long'] = 1

    def populate_indicators(self, dataframe: pd.DataFrame, metadata: dict) -> pd.DataFrame:

        # Ichimoku calculations
        # Conversion line
        high = dataframe['high'].rolling(9).max()
        low = dataframe['low'].rolling(9).min()
        dataframe['conversion'] = (high + low) / 2

        # Base line
        high2 = dataframe['high'].rolling(26).max()
        low2 = dataframe['low'].rolling(26).min()
        dataframe['base'] = (high2 + low2) / 2

        # Leading Span A
        dataframe['span_a'] = (dataframe['conversion'] + dataframe['base']) / 2

        # Leading Span B
        high3 = dataframe['high'].rolling(52).max()
        low3 = dataframe['low'].rolling(52).min()
        dataframe['span_b'] = (high3 + low3) / 2

        # Lagging Span
        dataframe['lagging'] = dataframe['close'].shift(-26)

        # Simple Moving Average
        dataframe['sma'] = pta.sma(dataframe['close'], timeperiod = 14)

        # Relative Strength Index
        dataframe['rsi'] = pta.rsi(dataframe['close'], length=14)
        
        # Compare the two span lines and select the one with the lowest value
        dataframe['stoploss_line'] = dataframe[['span_a', 'span_b']].min(axis=1)

        return dataframe

    def populate_entry_trend(self, dataframe: pd.DataFrame, metadata: dict) -> pd.DataFrame:

        dataframe.loc[
            (
                # Enter long when volume > 0
                (dataframe['volume'] > 0) &
                # Enter long when close price is above the cloud
                (dataframe['close'] >= dataframe['span_a']) &
                (dataframe['close'] >= dataframe['span_b']) &
                # Enter long when cloud is green
                (dataframe['span_a'] > dataframe['span_b']) &
                # Enter long when the conversion line is above the base line
                (dataframe['conversion'] > dataframe['base']) & 
                # Enter long when the lagging line is above the cloud
                (dataframe['lagging'] >= dataframe['span_a']) &              
                # Enter long when span_a crosses_above span_b
                (qtpylib.crossed_above(dataframe['span_a'], dataframe['span_b']))
            ),
            'enter_long'] = 1
        return dataframe

    def populate_exit_trend(self, dataframe: pd.DataFrame, metadata: dict) -> pd.DataFrame:

        # dataframe.loc[
            # (
            #     # Exit trade when volume > 0
            #     (dataframe['volume'] > 0) &
            #     # Exit trade when the lagging line goes below the price
            #     (qtpylib.crossed_below(dataframe['lagging'], dataframe['conversion']))
            # ),
            # 'exit_long'] = 1
        return dataframe
